#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
  run.py 21/1/2009 v. 0.4
  Copyright (C) 2008, 2009, Davide Lasagna
  Author: Davide Lasagna                 Email: lasagnadavide@gmail.com

   Distributed under the terms of the GNU General Public License. You are
  free to use/modify/distribute this program as long as you comply to the
    terms of the GNU General Public License, version 2 or above, at your
      option, and provided that this copyright notice remains intact.
"""
import gobject
import gtk.glade
import os
import string
import sys 
import re
import find_matches as fm
import df

class rundialog:
	def __init__(self):
		#Load the glade interface and get the entry widget
		if len(sys.argv) > 1:
			interface = sys.argv[1][sys.argv[1].rfind('=')+1:]
		else:
			interface = 'slim'
		
		self.gladefile = os.path.dirname(os.path.abspath(sys.argv[0])) +"/%s.glade"%interface
		
		self.wTree = gtk.glade.XML(self.gladefile)
		self.entry = self.wTree.get_widget("entry")
		self.matches_label = self.wTree.get_widget("matches_label")
		self.notification_label = self.wTree.get_widget("notification_label")
		self.program_icon = self.wTree.get_widget("program_icon")
		
		#Connect signals to actions
		dic = {  "on_Run_clicked"						: self.exec_command_line, \
					"on_Cancel_clicked"					: gtk.main_quit, \
					"on_terminal_toggled"				: self.terminal, \
					"on_root_toggled"						: self.root, \
					"on_entry_activate"					: self.exec_command_line, \
					"on_window_delete_event"			: gtk.main_quit }
		self.wTree.signal_autoconnect(dic)
		
		self.handle0 = self.entry.connect_after("key_press_event", self.key_press_event)
		self.handle1 = self.entry.connect("notify::cursor-position", self.update_matches_in_commands)
		self.handle2 = self.entry.connect("notify::cursor-position", self.update_matches_in_history)
		self.handle3 = self.entry.connect_after("notify::cursor-position", self.update_icon)
		self.handle4 = self.entry.connect_after("notify::cursor-position", self.update_notification_label)
		self.handle5 = self.entry.connect_after("notify::cursor-position", self.update_matches_label)
		
		#Create a list from history in ~/.bash_history	file
		self.history = []
		self.history_file = (os.getenv('HOME')+'/.run_history')
		#check if history file exixts and if not create a new one
		if os.access(self.history_file,os.F_OK):
			f = open(self.history_file)
		else:
			os.popen("touch %s"%self.history_file)
			f = open(self.history_file)
		#load history
		for line in f:
			self.history.append(line[0:-1])
		#then reverse it, cause last has to be first
		self.history.reverse()
		#remove dups
		outlist = []
		for element in self.history:
			if element not in outlist:
				outlist.append(element)
		self.history = outlist
		
		#Load possible commands found in $PATH
		self.commands = []
		Paths =  string.split(os.getenv('PATH'),sep=':')
		for path in Paths:
			binaries = os.listdir(path)
			self.commands.extend(binaries)
		self.commands.sort()
		
		#some initial stuff
		self.run_in_terminal = False
		self.run_as_root = False
		self.pointer = -1
		self.hist_pointer = 0
		
		#get iconset
		self.iconset = gtk.icon_theme_get_default()
		#get desktop files
		self.d_files = df.desktop_files()
				
		#Update matches and gui
		self.update_matches_in_commands()
		self.update_matches_in_history()
		self.update_icon()
		self.update_notification_label()
		self.update_matches_label()
		

	def key_press_event(self,widget,event): #how to handle special keys....
		#print "run key_press_event"
		if (event.keyval == gtk.keysyms.Page_Up or event.keyval == gtk.keysyms.Up) and self.hist_matches: 
			#Increase by one the pointer locating the current history item
			self.hist_pointer = self.hist_pointer + 1
			if self.hist_pointer >= len(self.hist_matches) - 1:
				self.hist_pointer = len(self.hist_matches) - 1
				self.code = 'tophistory'
				self.update_notification_label()
				self.code = 'ok'
			#print self.hist_pointer
			#Set entry Text and cursor position
			self.entry.handler_block(self.handle2)
			self.entry.handler_block(self.handle1)
			pos = self.entry.get_position()
			self.entry.set_text(self.hist_matches[self.hist_pointer])
			self.entry.set_position(pos)
			self.entry.handler_unblock(self.handle2)
			self.entry.handler_unblock(self.handle1)
			self.update_icon()
			return True
	
		if (event.keyval == gtk.keysyms.Page_Down or event.keyval == gtk.keysyms.Down) and self.hist_matches: 
			#Decrease by one the pointer locating the current history item
			#limit is zero (just prefix)
			self.hist_pointer = self.hist_pointer - 1
			if self.hist_pointer < 0:
				self.hist_pointer = 0
			#print self.hist_pointer
			#Set entry Text and cursor position
			self.entry.handler_block(self.handle2)
			self.entry.handler_block(self.handle1)
			pos = self.entry.get_position()
			self.entry.set_text(self.hist_matches[self.hist_pointer])
			self.entry.set_position(pos)
			self.entry.handler_unblock(self.handle2)
			self.entry.handler_unblock(self.handle1)
			self.update_icon()
			return True

		if event.keyval == gtk.keysyms.Escape:
			gtk.main_quit()

		if event.state & gtk.gdk.CONTROL_MASK and event.keyval == gtk.keysyms.Tab:
			self.pointer = self.pointer -1
			if self.pointer < 0:
				self.pointer = len(self.comm_matches)-1
				
			if self.comm_matches:
				self.entry.handler_block(self.handle1)
				self.entry.handler_block(self.handle2)
				self.entry.handler_block(self.handle3)
				self.entry.handler_block(self.handle4)
				self.entry.handler_block(self.handle5)
				pos = self.entry.get_position()
				self.entry.set_text(self.comm_matches[self.pointer])
				self.entry.set_position(pos)
				self.entry.handler_unblock(self.handle1)
				self.entry.handler_unblock(self.handle2)
				self.entry.handler_unblock(self.handle3)
				self.entry.handler_unblock(self.handle4)
				self.entry.handler_unblock(self.handle5)
				self.update_icon()
			return True
			
		if event.keyval == gtk.keysyms.Tab:
			self.pointer = self.pointer +1
			if self.pointer >= len(self.comm_matches):
				self.pointer = 0
			if self.comm_matches:
				self.entry.handler_block(self.handle1)
				self.entry.handler_block(self.handle2)
				self.entry.handler_block(self.handle3)
				self.entry.handler_block(self.handle4)
				self.entry.handler_block(self.handle5)
		
				
				pos = self.entry.get_position()
				self.entry.set_text(self.comm_matches[self.pointer])
				if len(self.comm_matches) == 1:
					new_pos = re.search(r'(?<!\\) ',self.entry.get_text()[pos:])
					if new_pos:
						print new_pos.end()
						if new_pos.end() == 0:
							new_pos = -1
						else:
							new_pos = new_pos.end()+2
					else:
						new_pos = -1
					self.entry.set_position(new_pos)
					self.update_matches_in_commands()
					self.update_matches_in_history()
				else:
					self.entry.set_position(pos)
					
				self.entry.handler_unblock(self.handle1)
				self.entry.handler_unblock(self.handle2)
				self.entry.handler_unblock(self.handle3)
				self.entry.handler_unblock(self.handle4)
				self.entry.handler_unblock(self.handle5)
				
				self.update_icon()
				self.update_matches_label()
				self.update_notification_label()
				
			return True
			
		if event.state  &  (gtk.gdk.CONTROL_MASK | gtk.gdk.SHIFT_MASK) and event.keyval == gtk.keysyms.Return:
			if event.state & gtk.gdk.SHIFT_MASK and event.state & gtk.gdk.CONTROL_MASK:
					self.run_as_root = True;	self.run_in_terminal = True
					self.exec_command_line(self)
					return True
			if event.state & gtk.gdk.CONTROL_MASK:
					self.run_in_terminal = True;
					self.exec_command_line(self)
					return True
			if event.state & gtk.gdk.SHIFT_MASK:
					self.run_as_root = True;
					self.exec_command_line(self)
					return True
		return False

	def update_matches_in_commands(self,*args): #Get commands in PATH matching with a prefix 
		#print "run update_matches_in_commands"
		#get text and cursor position
		entry_text = self.entry.get_text()
		curs_pos = self.entry.get_position()
		#reset pointer to list elements
		self.hist_pointer = 0
		self.pointer = -1
		  

		#so i have my_text, now i have to split it in the part before the cursor,
		#which is going to be completed, and the following part, which doesnt matter.
		text_tobe_completed = fm.find_text_tobe_completed(entry_text,curs_pos)
		
		#we want to find matches. These can be either pathnames or programs.
		if text_tobe_completed and text_tobe_completed[0] == '/':
			matches,self.code, self.matches_number = fm.find_path_matches(text_tobe_completed)
		else:
			#this is a list of programs matching with the text given
			matches,self.code, self.matches_number = fm.find_comm_matches(text_tobe_completed,self.commands)

		#now complete the whole line
		pre_slice = entry_text[0:curs_pos]
		post_slice = entry_text[curs_pos:]

		pre_temp  = re.search(r" (?!\\)",pre_slice[::-1])
		post_temp = re.search(r"(?<!\\) ",post_slice)
		if post_temp: 
			post_idx = curs_pos + post_temp.end()
			if entry_text[post_idx-2] == '\\':
				post_idx = -1
		else:
			post_idx = -1
		if pre_temp: 
			pre_idx  = curs_pos - pre_temp.start()
		else:
			pre_idx = 0
		
		text_matches = []
		#~ if post_idx == -1:
			#~ text_matches = [entry_text[0:curs_pos]]
		#~ else: 
			#~ text_matches = [entry_text[0:curs_pos]+' '+entry_text[post_idx:]]
			
		for mat in matches:
			if post_idx == -1:
				completion_string = entry_text[0:pre_idx]+mat
			else:
				completion_string = entry_text[0:pre_idx]+mat+' '+entry_text[post_idx:]
			text_matches.append(completion_string)
			
		self.comm_matches = text_matches
		return False


	def update_matches_in_history(self,*args): #Get commands in history matching with a prefix 
		#print "run update_matches_in_history"
		#get text and cursor position
		entry_text = self.entry.get_text()
		curs_pos = self.entry.get_position()
		prefix = entry_text[0:curs_pos]
		self.hist_pointer = 0

		self.hist_matches = [prefix]
		for comm in self.history:
			if comm.startswith(prefix) and comm != '':
				self.hist_matches.append(comm)
		return False


	def exec_command_line(self,*args):#Run the specified command	
		#get the command the user wants to run
		comm = self.entry.get_text()

		#start considering flags
		stringa = comm
		#print 
		#print "comm = ", comm
		
		#handle opening files with default application
		if os.access(re.sub(r'(\\ )',r' ',comm),os.F_OK) or comm.startswith('http'):
			stringa = 'xdg-open "' + re.sub(r'(\\ )',r' ',comm) + '"'
			#print "opening with default application"

		#handle control and shift keys
		if self.run_in_terminal == True:
			if self.run_as_root == True:
				stringa = 'gksu "gnome-terminal --execute %s"'%stringa
			else:
				stringa = 'gnome-terminal --execute %s'%stringa
		else:
			if self.run_as_root == True:
				stringa = 'gksu "%s"'%stringa

		#make the program run in the background
		stringa = stringa+" &"
		#print 
		#print "st = ", stringa
		
		#Append entered command in history file 
		os.system('echo %s >> %s'%(comm,self.history_file))
		try:
			comm_output = os.popen2(stringa)
		except:
			#~ print "error running command"
			pass
		gtk.main_quit()


	def update_matches_label(self,*args):
		#print "run update_matches_label"
		if self.matches_number == 1:
			self.matches_label.set_markup('<span weight="bold">1</span> match found')
		else:
			self.matches_label.set_markup('<span weight="bold">%d</span> matches found'%( self.matches_number ))
		return False
	def update_notification_label(self,*args):
		#print "run update_notification_label"
		#i want to write a little suggestion in the label. This function handles different situations
		if self.code == "tophistory":
			self.notification_label.set_markup('<span weight="bold" foreground="#B44A35">Top of history</span>')
			gobject.timeout_add(2000,lambda:self.notification_label.set_markup('<span weight="bold">Insert Command</span>'))
			return
		if self.code == 'notfound' or self.matches_number == 0:
			self.notification_label.set_markup('<span weight="bold" foreground="#B44A35">No completion possible!</span>')
			gobject.timeout_add(2000,lambda:self.notification_label.set_markup('<span weight="bold">Insert Command</span>'))
			return
		if self.code == 'accessdenied':
			self.notification_label.set_markup('<span weight="bold" foreground="#B44A35">Cannot change directory</span>')
			gobject.timeout_add(2000,lambda:self.notification_label.set_markup('<span weight="bold">Insert Command</span>'))
			return
		if self.matches_number==1:
			self.notification_label.set_markup('<span weight="bold" foreground="#35B45C">Pattern unique</span>')
			return
		if self.entry.get_position() != 0:
			self.notification_label.set_markup('<span weight="bold">Press tab to complete</span>')
			return
		if self.entry.get_position() == 0:
			self.notification_label.set_markup('<span weight="bold">Insert Command</span>')
			return
		return False
	
	def terminal(self,*args): # Toggle "run in terminal" options from GUI checkbox
		if self.run_in_terminal: self.run_in_terminal = False
		else: self.run_in_terminal = True
		self.entry.handler_block(self.handle1)
		self.entry.handler_block(self.handle2)
		self.entry.handler_block(self.handle3)
		self.entry.handler_block(self.handle4)
		self.entry.handler_block(self.handle5)
		pos = self.entry.get_position()
		self.entry.grab_focus()
		self.entry.set_position(pos)
		self.entry.handler_unblock(self.handle1)
		self.entry.handler_unblock(self.handle2)
		self.entry.handler_unblock(self.handle3)
		self.entry.handler_unblock(self.handle4)
		self.entry.handler_unblock(self.handle5)
		
		#~ print "terminal toggled", self.run_in_terminal

	def root(self,*args):# Toggle "run as root" options from GUI checkbox
		if self.run_as_root: self.run_as_root = False
		else:	self.run_as_root = True
		self.entry.handler_block(self.handle1)
		self.entry.handler_block(self.handle2)
		self.entry.handler_block(self.handle3)
		self.entry.handler_block(self.handle4)
		self.entry.handler_block(self.handle5)
		pos = self.entry.get_position()
		self.entry.grab_focus()
		self.entry.set_position(pos)
		self.entry.handler_unblock(self.handle1)
		self.entry.handler_unblock(self.handle2)
		self.entry.handler_unblock(self.handle3)
		self.entry.handler_unblock(self.handle4)
		self.entry.handler_unblock(self.handle5)
	def update_icon(self,*args):
		#print "run update_icon"
		#when needed this function finds the the best icon related to
		#the program to run. If it fails it shows a fallback icon
		comm = self.entry.get_text()
		#we try to use the icon suggested in the desktop file of 
		#the program we want to run
		icon_name = self.d_files.from_command_get_icon_name(comm )
		#if this exists
		if icon_name:
			#if it is a filename
			if os.access(icon_name,os.F_OK):
				self.program_icon.set_from_file(icon_name)
				#print "icon is file"
			else:
				self.program_icon.set_from_icon_name(icon_name,6)
				#print "icon has a strange name but is in theme"
		#if there is no icon specified in the desktop file
		#we try if the icon theme has this icon
		else:
			if self.iconset.has_icon(comm):
				self.program_icon.set_from_icon_name(comm,6)
				#print "icontheme has icon"
			#otherwise we use the fall back icon
			else:
				self.program_icon.set_from_icon_name('gnome-run',6)
				#print "azz"
		
		return
				

app=rundialog()
gtk.main()
















